(lambda __g: (lambda __mod: [(setup(name='oruline', version='1.0.2', author='Ganesh Kumar T K', author_email='ganeshkumartk@outlook.com', maintainer='Ganesh', maintainer_email='ganeshkumartk@outlook.com', license='MIT', url='https://github.com/coderganesh/1line', description='Convert any Python file into a single line of code.', long_description=open('README.md', 'r').read(), long_description_content_type='text/markdown', packages=['oruline'], keywords = ['obfuscate', 'oneliner', 'short','oneline','single','oruline','lambda'], classifiers=['Development Status :: 5 - Production/Stable','Intended Audience :: Developers','Topic :: Software Development :: Build Tools','License :: OSI Approved :: MIT License','Programming Language :: Python :: 3','Programming Language :: Python :: 3.4','Programming Language :: Python :: 3.5','Programming Language :: Python :: 3.6','Programming Language :: Python :: 3.7'], entry_points={'console_scripts': ['oruline=oruline.__main__:main']}, test_suite='oruline.runtests'), None)[1] for __g['setup'] in [(__mod.setup)]][0])(__import__('setuptools', __g, __g, ('setup',), 0)))(globals())
